//Boston University - Electronics Design Facility
//Monkey Timer Mk2
//http://edf.bu.edu/
//written by Chris Lawlor, January - April 2011
//Device is Atmega2560
//CPU clock is 16Mhz
//This file defines the interface for the LSI NT3881 LCD display controller


#ifndef _MONKEY_MK2_DISPLAY_H
#define _MONKEY_MK2_DISPLAY_H 1

#include <string.h>


#define BACKLIGHT	0x02	


#define CONTROL		PORTF
#define	CTRL_MSK	0xE0
#define CTRL_RW		0x40	//these are the bitmasks for the respective signal in port F
#define CRTL_E		0x80	//
#define CRTL_RS		0x20	//


#define DB_PORT_DDR	DDRK
#define	DB_PORT		PORTK
#define DB_PIN		PINK


//display commands
//for these ones RS = 0 and RW = 0
//ie WRITE COMMANDS to control registers
#define DISPLAY_CLEAR		0x01	//1.64ms  Clear entire display area, restore display from shift, and load address counter with DD RAM address 00H.

#define DISPLAY_CHOME		0x02	//1.64ms  Restore display from shift and load address counter with DD RAM address 00H.

#define DISPLAY_SEMODE_MSK	0x04	//40us Specify direction of cursor movement and display shift mode. This operation takes place after each data transfer (read/write).
#define DISPLAY_SEMODE_FWD	0x02	//this bit controlls direction
#define DISPLAY_SEMODE_SHFT	0x01	//this bit controlls shift

#define DISPLAY_ONOFF_MSK	0x08	//40us Specify activation of display (D) cursor (C) and blinking of character at cursor position (B).
#define DISPLAY_ONOFF_DSPLY	0x04	//display activation	
#define DISPLAY_ONOFF_CRSR	0x02	//shows the cursor
#define DISPLAY_ONOFF_BLNK	0x01	//curser blinking	

#define DISPLAY_DCS_MSK		0x10	//40us Shift display or move cursor.
#define DISPLAY_DCS_SC		0x08	//?
#define DISPLAY_DCS_RL		0x04	//?

#define DISPLAY_FUNC_MSK	0x20	//40us Set interface data length (DL), number of display line (N), and character font (F).
#define DISPLAY_FUNC_DLEN	0x10	//
#define DISPLAY_FUNC_NDISP	0x08	//
#define DISPLAY_FUNC_FONT	0x04	//

#define DISPLAY_RAM_ADDR	0x40	//40us Load the address counter with a CG RAM address. Subsequent data access is for CG RAM data.
									//address goes in the lower 6 bits

#define DISPLAY_DDRAM_ADST	0x80	//40us Load the address counter with a DD RAM address. Subsequent data access is for DD RAM data.
									//address goes in the lower 7 bits


//for these ones RS = 0 and RW = 1	
//ie READ COMMANDS from control registers
#define DISPLAY_BSYFLAG_MSK	0x80	//40us Load the address counter with a DD RAM address. Subsequent data access is for DD RAM data.
#define DISPLAY_ADRCNT_MSK	0x7F	

	
//for these ones RS = 1 and RW = 0
//ie write COMMANDS to data memory
#define DISPLAY_DWRITE_MSK	0xFF	//40us Write data to CG RAM or DD RAM.

//for these ones RS = 1 and RW = 1
//ie read COMMANDS from data memory
#define DISPLAY_DREAD_MSK	0xFF	//40us Read data from CG RAM or DD RAM.


//#define DISPLAY_TEST_STR	"023456789*123456789*223456789*323456789*423456789*523456789*623456789*723456789*823456789*923456789*A23456789*B23456789*C23456789*D23456789*E23456789*F23456789*"


#define	NO_OP	__asm__ __volatile__( "nop" ) 	//each nop is 63 ns (ie, 1 cpu clock cycle)


#define DISPLAY_USES_ISR

#ifdef DISPLAY_USES_ISR

	#define LONG_TIME	41
	#define SHORT_TIME	2

	volatile uint16_t	display_lockout_timer = 0;	//-1 indicated lockout is disabled (off), >=0 indicates enabled

	uint8_t	display_set_lockout(uint16_t n)
	{
		if(display_lockout_timer)
			return -1;
		else
		{
			display_lockout_timer = n;
			//setup timer counter
			TCCR0A = 0x02;
			OCR0A = 10;
			TIMSK0 = 0x02;		//no interrupt used.
			TCCR0B = 0x03;
			return 1;
		}
	}


	ISR(TIMER0_COMPA_vect, ISR_BLOCK)
	{
		if (display_lockout_timer == 1)
		{
			TCCR0B = 0x00;		//disable the TIMER0_COMPA interrupt
			TIMSK0 = 0x00;
			//for asynchronous display mode, do a software interrupt here as a signal
		}
		display_lockout_timer--;	//decriment display_lockout_timer, 0 will indicate off
	}


	void display_lockout_wait()
	//sets display_lockout_timer to 1 until n*40us have elapsed
	{
		while (display_lockout_timer)
		{}
	}

#else
	//in non ISR mode there's a screen garbling issue durring the run_trial()
	//the screen seems to shift by +8 chars when starting the beam
	//ideas: 
	//probably: the beam interrupts are interfering durring the run_trial()
	//or there's an error in the display_menu stuff

	#define LONG_TIME	35
	#define SHORT_TIME	1

	void display_wait(uint8_t n)
	//waits for n counts of 80us (the screen needs at least 40us)
	//for a command that needs 40us n should be 1
	//for a command that needs 1.64ms n should be 35
	{
		//setup timer
		TCCR0A = 0x02;
		OCR0A = 20;
		TIMSK0 = 0x00;
		TCCR0B = 0x03;	

		//block and poll
		while (n>0)
		{
			while(!(TIFR0&(1<<OCF0A)))
			{}
			TIFR0 |= (1<<OCF0A);
			n--;
		}
		//disable timer
		TCCR0B = 0x00;
	}
#endif


void display_block(uint8_t t)
{
	#ifdef DISPLAY_USES_ISR
		display_set_lockout(t);
		display_lockout_wait();
	#else
		display_wait(t);
	#endif
}


void display_write(uint8_t data, uint8_t rsbit)
//this NO_OP delay technique relies on the knowing CPU frequency
{
	CONTROL &= ~CTRL_MSK;			//drive E low, set RW low for write

	if (rsbit)
		CONTROL |= 	CRTL_RS;		//set RS bit

	NO_OP;							//wait >60ns, (each nop is 63 ns)
	
	CONTROL |= CRTL_E;				//drive E high
	
	DB_PORT = data;					//write data byte
	
	NO_OP;NO_OP;NO_OP;				//wait >300ns
	NO_OP;NO_OP;NO_OP;				
	
	CONTROL &= ~CRTL_E;				//drive E low
	
	NO_OP;							//wait >10ns
	
	//CONTROL |= CTRL_RW;			//set RW to high (this is not necessary)

	NO_OP;NO_OP;NO_OP;				//pad out the cycle time
	NO_OP;NO_OP;NO_OP;				//(this is probably not necessary as long as the display_lockout_timer is being used)
}



void display_clear(uint8_t noblock)
//Clear entire display area, restore display from shift, and load address counter with DD RAM address 00H.
//a noblock value of 1 will cause the function to return without waiting for the display_lockout_timer to expire
//a value of 0 will cause it to block execution until the display_lockout_timer expires (1.64ms)
{
	uint8_t db = DISPLAY_CLEAR;

	display_write(db, 0);

	
	if (!noblock)
	{
		display_block(LONG_TIME);
	}
}

void display_home(uint8_t noblock)
//a noblock value of 1 will cause the function to return without waiting for the display_lockout_timer to expire
//a value of 0 will cause it to block execution until the display_lockout_timer expires
{
	uint8_t db = DISPLAY_CHOME;

	display_write(db, 0);

	if (!noblock)
	{
		display_block(LONG_TIME);
	}

}

void display_entrymode(uint8_t inc, uint8_t shift, uint8_t noblock)
//an inc value of 1 will incriment the curser, a value of 0 will decriment the curser
//a shift value of 1 will result in shift, 0 will be no shift
//a noblock value of 1 will cause the function to return without waiting for the display_lockout_timer to expire
//a value of 0 will cause it to block execution until the display_lockout_timer expires
{
	uint8_t db = DISPLAY_SEMODE_MSK;
	
	if (inc)
		db |= DISPLAY_SEMODE_FWD;
		
	if (shift)
		db |= DISPLAY_SEMODE_SHFT;

	display_write(db, 0);

	if (!noblock)
	{
		display_block(SHORT_TIME);
	}
}


void display_onoff(uint8_t dsply, uint8_t curser, uint8_t blink, uint8_t noblock)
//a dsply value of 1 will turn the display on, 0 for off
//a curser value of 1 will display the curser, a 0 will hide it
//a blink value of 1 will blink the curser, a 0 will make it not blink
//a noblock value of 1 will cause the function to return without waiting for the display_lockout_timer to expire
//a value of 0 will cause it to block execution until the display_lockout_timer expires
{
	uint8_t db = DISPLAY_ONOFF_MSK;
	
	if (dsply)
		db |= DISPLAY_ONOFF_DSPLY;
		
	if (curser)
		db |= DISPLAY_ONOFF_CRSR;

	if (blink)
		db |= DISPLAY_ONOFF_BLNK;

	display_write(db, 0);

	if (!noblock)
	{
		display_block(SHORT_TIME);
	}
}

void display_function_set(uint8_t data_length, uint8_t num_lines, uint8_t font, uint8_t noblock)
//Sets interface data length (DL), number of display lines (N), and character font (F).
//a data_length value of 1 selects 8-Bit, 0 selects 4-Bit
//a num_lines value of 1 selects Dual Line, a 0 selects Signal Line
//a font value of 1 selects 5x10 dots, a 0 selects 5x8 dots
//a noblock value of 1 will cause the function to return without waiting for the display_lockout_timer to expire
//a value of 0 will cause it to block execution until the display_lockout_timer expires
{
	uint8_t db = DISPLAY_FUNC_MSK;
	
	if (data_length)
		db |= DISPLAY_FUNC_DLEN;
		
	if (num_lines)
		db |= DISPLAY_FUNC_NDISP;

	if (font)
		db |= DISPLAY_FUNC_FONT;

	display_write(db, 0);

	if (!noblock)
	{
		display_block(SHORT_TIME);
	}
}


void display_curser_shift(uint8_t shift_display, uint8_t right, uint8_t noblock)
//a shift_display value of 1 shifts the display, a value of 0 moves the curser
//a right value of 1 shifts right, a value of 0 shifts left
//a noblock value of 1 will cause the function to return without waiting for the display_lockout_timer to expire
//a value of 0 will cause it to block execution until the display_lockout_timer expires
{
	uint8_t db = DISPLAY_DCS_MSK;
	
	if (shift_display)
		db |= DISPLAY_DCS_SC;

	if (right)
		db |= DISPLAY_DCS_RL;

	display_write(db, 0);

	if (!noblock)
	{
		display_block(SHORT_TIME);
	}

}


void display_data_write(uint8_t data, uint8_t noblock)
//writes a byte to the display's DD RAM
//a noblock value of 1 will cause the function to return without waiting for the display_lockout_timer to expire
//a value of 0 will cause it to block execution until the display_lockout_timer expires
{
	display_write(data, 1);

	if (!noblock)
	{
		display_block(SHORT_TIME);
	}
}


void display_string(char *str)
//displays a character string on the display
{
	char nl[] = "(null)";
	char *c = nl;

	if (str)
		c = str;
	
	//USART_SendString(str, 0);
	//USART_SendLineFeed();	

	while(*c != '\0')
	{
		display_data_write(*c, 0);	
		display_entrymode(1,0,0);
		c++;
	}
}

/*
void display_test_str_shift()
{
	display_string(DISPLAY_TEST_STR);
	while(1)
	{
		display_curser_shift(1,0,0);
		msleep(1000);
	}
}
*/

#define ROWS	4
#define COLS	20							//
#define NSEQ	1							//can make it bigger as size permits

typedef  char screen_t [ROWS][COLS+1];		//[row][col]	the +1 is for the \0 null termination character
screen_t 	screen[NSEQ]; 					//screen[i][j][k] --->  screen[seqnum][row][col]

//screen_t	testscreen1 = {,,,,};
//static uint16_t	seqnum = 0;						//this global var keeps track of the current screen

#define MIN(x,y) 	(((x)<(y))?(x):(y))


void display_init()
{
	DDRC  |= BACKLIGHT;	
	PORTC |= BACKLIGHT;	

	DDRF |= CTRL_MSK;		//for the control signal line
	DDRK |= 0xFF;			//for the data lines

	memset(screen,0,NSEQ*sizeof(screen_t));

	//display_set_lockout(4);	//wait a startup time.  41 works (1.6ms), 2 works (80us)
	//display_lockout_wait();
	//display_wait(4);
	display_block(2*SHORT_TIME);

	display_onoff(1,0,0,0);	//display on, curser off, blink off
	display_clear(0);
	display_home(0);		
	//display_entrymode(1,0,0);
	display_function_set(1,1,0,0);
	//display_test_str_shift();
	//display_string(DISPLAY_TEST_STR);
}



//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------





void blank_screen(screen_t *scr)
//fills a screen_t with all spaces
{
	for (uint8_t i = 0; i<ROWS; i++)
	{
		memset( (*scr)[i], ' ', COLS);
	}
}

uint8_t screen_index_permutation(uint8_t i)
//returns the screen permutation of an index
{
	switch (i)
	{
		case 0:
			return 0;
			break;
		case 1:
			return 2;
			break;
		case 2:
			return 1;
			break;
		case 3:
			return 3;
			break;
		default:
			return i;
			break;
	};
}

#define SIP(x)	(screen_index_permutation((x)))

/*
void USART_SendScreen(screen_t *scr)
//sends a screen_t over USART
{
	char m[32];
	char n[32];

	for (int i = 0; i<ROWS; i++)
	{
		memset(m,0,32);
		memset(n,0,32);
		strncpy(m, (*scr)[i], COLS);
		snprintf(n,32,"%u<%s>\n",i,m);
		USART_SendString(n,0);
		//USART_SendString( "<", 0);
		//USART_SendString( (*scr)[i], COLS);
		//USART_SendString( ">\n", 0);
	}

}
*/

void display_str2scr(screen_t *scr, char *s1, char *s2, char *s3, char *s4)
//the general idea is that each sN displays on line N
//if sN is too long the line is truncated on the display
//
//
//  //--This comment does not currently apply--
//  //--in the future maybe text fitting will be implemented--
//  //the caveat is that they all have to be less than 20chars in length
//  //so, a word wrapping shceme is employed, as follows
//  //if any of the sN pointers is null, then that line is omitted
//  //and is used as an overflow line
//  //The total length of all of the strings should not exceed 80 chars
{
	char *s[ROWS];
	s[0] = s1;
	s[1] = s2;
	s[2] = s3;
	s[3] = s4;

	uint8_t l[ROWS] ;
	memset(l, 0, ROWS);

	uint8_t d[ROWS] ;
	memset(d, 0, ROWS);

	blank_screen(scr);

	for(int i=0; i<ROWS; i++)
	{
		if (s[i])
			l[i] = strlen(s[i]);
		d[i] = MIN(l[i],COLS);
	}

	//maybe could do some text fitting analysis here?
	
	for(int i=0; i<ROWS; i++)
	{
		if (d[i])	
		{
			strncpy( (*scr)[i], s[i], d[i]);
		}		
		(*scr)[i][COLS] = '\0';
	}

	//USART_SendScreen(scr);			//for testing
}


void display_screen(screen_t *scr)
//takes a screen_t and displays it on the screen
{
	//display_clear(0);	//this isn't necessary if we always write 80 chars at a time 
	//display_home(0);	//this isn't necessary either for the sane reason however, the timing looks nicer
	//USART_SendString("display_screen():  \n", 0);
	//USART_SendScreen(scr);

	//check for early '\0' charachers
	//this prevents a lot of potential heartache caused by an early terminator '\0' character garbling up the display
	
	for (uint8_t j = 0; j<ROWS; j++)
	{
		uint8_t found_null = 0;
		for (uint8_t k = 0; k<COLS; k++)
		{
			if ( (found_null==1) || ( (*scr)[j][k] == '\0') )
			{
				(*scr)[j][k] = ' ';
				found_null = 1;
			}
		}
		(*scr)[j][COLS] = '\0';	//just in case
	}


	for (uint8_t i=0; i<ROWS; i++)
	{
		display_string( (*scr)[SIP(i)] );
	}

	//USART_SendString("display_screen():  finished\n", 0);
}

/*
void display_msg(char *s1, char *s2, char *s3, char *s4)
//the general idea is that each sN displays on line N
//if sN is too long the line is truncated on the display
{
	//volatile screen_t temp;
	display_str2scr(screen, s1, s2, s3, s4);
	display_screen(screen);
}
*/
/*
void screen_test()
{
	char t1[] = "This is the first";
	char t2[] = "This is the second";
	char t3[] = "This is the third";
	char t4[] = "This is the fourth";
	
	display_str2scr(screen, t1, t2, t3, t4);
	display_screen(screen);
}
*/

#endif
