//Boston University - Electronics Design Facility
//Monkey Timer Mk2
//http://edf.bu.edu/
//written by Chris Lawlor, January - April 2011
//Device is Atmega2560
//CPU clock is 16Mhz
//This file defines the interface for the Maxim DS1307 Real Time Clock 
//the Maxim DS1307 uses the TWI at 100kHz

#ifndef _RTC_DS1307_H
#define _RTC_DS1307_H 1

#include <avr/interrupt.h>
#include "twi_atmega2560.h"
#include "monkey_mk2_datatypes.h"

//DS1307 REGISTERS
#define DS1307_SECONDS 		0x00
#define DS1307_MINUTES 		0x01
#define DS1307_HOURS 		0x02
#define DS1307_DAY 			0x03
#define DS1307_DATE 		0x04
#define DS1307_MONTH 		0x05
#define DS1307_YEAR 		0x06
#define DS1307_CONTROL 		0x07
#define DS1307_RAM_BOTTOM	0x08
//#define DS1307_RAM_START	0x09
#define DS1307_IS_INIT		0x3F	//a custom one:  initialized to a known value to check initialization
#define DS1307_RAM_TOP		0x3F

//These are the initialization values for the DS1307 REGISTERS
#define SEC_VALUE			0x00		
#define MIN_VALUE			0x00
#define HRS_VALUE			0x00
#define DAY_VALUE			0x01
#define DATE_VALUE			0x01
#define MNTH_VALUE			0x01
#define YEAR_VALUE			0x00
#define CONTROL_VALUE		0x03	//
#define	INIT_VALUE			0x75

//These are some other important things
#define DS1307_TWI_ADDR		0x68	//The TWI Slave address of the DS1307
#define DS1307_SQW_FREQ		32768	//the actual frequency is set by CONTROL_VALUE, the number here must match the setting
#define SQW_PIN						//PORTJ pin 0

uint8_t ds1307_is_init()
{
	//read the DS1307_IS_INIT byte from the RAM
	uint8_t iv = 0;
	twi_master_read(DS1307_TWI_ADDR, DS1307_IS_INIT, (char*)&iv, 1); 
	return (iv == INIT_VALUE);
}

uint8_t ds1307_init()
//initializes the RTC and checks that it works
//(writes a 0x00 to the seconds register to enable the oscilator)
//afterword the date will be set to the initializatoin date (ie, whatever is referenced above)
//and the application will require a call to ds1307_set_date_time() to set the real date
//ds1307_init() only needs to be called once after the DS1307 RTC loses it's battery backup power, or on initial setup
//returns 1 on success or <0 on failure
{
	char init_vect[8] = {SEC_VALUE, MIN_VALUE, HRS_VALUE, DAY_VALUE, DATE_VALUE, MNTH_VALUE, YEAR_VALUE, CONTROL_VALUE};
	
	if ( twi_master_write(DS1307_TWI_ADDR, DS1307_SECONDS, init_vect, 8) < 8)
	{
		return -1;		
	}

	uint8_t iv = INIT_VALUE;	
	twi_master_write(DS1307_TWI_ADDR, DS1307_IS_INIT, (char*)&iv, 1);

	if (!ds1307_is_init())
	{
		return -2;
	}

	return 1;
}

/*
//for reference only

typedef struct a_time { 
	uint8_t month; 	// month 1-12
	uint8_t mday; 	// day 1-31
	uint8_t year; 	// year - 2000
	uint8_t hour; 	// hour 0-23
	uint8_t min; 	// minute 0-59
	uint8_t sec; 	// seconds 0-59
} a_time_t;
*/

#define BCD2BIN(bcd_byte, high_mask, low_mask)	(10*(((bcd_byte)&(high_mask))>>4)+((bcd_byte)&(low_mask)))
#define BIN2BCD(bin_byte, high_mask, low_mask)	(((((bin_byte)/10)<<4)&(high_mask))|(((bin_byte)%10)&(low_mask)))  


#define DS1307_SEC_HMSK		0x70
#define DS1307_SEC_LMSK		0x0F
#define DS1307_MIN_HMSK		0x70
#define DS1307_MIN_LMSK		0x0F
#define DS1307_HRS_HMSK		0x30
#define DS1307_HRS_LMSK		0x0F
#define DS1307_DAYW_HMSK	0x00
#define DS1307_DAYW_LMSK	0x07
#define DS1307_DATE_HMSK	0x30
#define DS1307_DATE_LMSK	0x0F
#define DS1307_MONTH_HMSK	0x10	
#define DS1307_MONTH_LMSK	0x0F	
#define DS1307_YEAR_HMSK	0xF0
#define DS1307_YEAR_LMSK	0x0F

enum{sec,min,hrs,day,date,month,yr};

void ds1307_read_date_time(a_time_t *t)
{
	char b[7] = {-1,-1,-1,-1,-1,-1,-1};

	twi_master_read(DS1307_TWI_ADDR, DS1307_SECONDS, b, 7);

	t->month	= BCD2BIN( b[month]	, DS1307_MONTH_HMSK, DS1307_MONTH_LMSK);
	t->mday		= BCD2BIN( b[date]	, DS1307_DATE_HMSK,  DS1307_DATE_LMSK); 
	t->year		= BCD2BIN( b[yr]	, DS1307_YEAR_HMSK,  DS1307_YEAR_LMSK);
	t->hour		= BCD2BIN( b[hrs]	, DS1307_HRS_HMSK,   DS1307_HRS_LMSK);
	t->min		= BCD2BIN( b[min]	, DS1307_MIN_HMSK,   DS1307_MIN_LMSK);
	t->sec		= BCD2BIN( b[sec]	, DS1307_SEC_HMSK,   DS1307_SEC_LMSK);

}

void ds1307_set_date_time(a_time_t *t)
{

	char b[7] = {-1,-1,-1,-1,-1,-1,-1};

	b[month] 	= BIN2BCD( t->month	, DS1307_MONTH_HMSK, DS1307_MONTH_LMSK);
	b[date] 	= BIN2BCD( t->mday	, DS1307_DATE_HMSK,  DS1307_DATE_LMSK); 
	b[yr] 		= BIN2BCD( t->year	, DS1307_YEAR_HMSK,  DS1307_YEAR_LMSK);
	b[hrs] 		= BIN2BCD( t->hour	, DS1307_HRS_HMSK,   DS1307_HRS_LMSK);
	b[min] 		= BIN2BCD( t->min	, DS1307_MIN_HMSK,   DS1307_MIN_LMSK);
	b[sec] 		= BIN2BCD( t->sec	, DS1307_SEC_HMSK,   DS1307_SEC_LMSK);

	twi_master_write(DS1307_TWI_ADDR, DS1307_SECONDS, b, 7);

}

	/*
	//this code does a multibyte write read for debug
	char b[10] = {0x0A,2,3,4,5,6,7,8,9,0xA};
	char c[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
	char tb[42] = "this is the test string\n";
	int n = twi_master_write(DS1307_TWI_ADDR, DS1307_RAM_START, b, 10);
	snprintf(tb, 41, "set():  wrote: %d bytes\n", n);
	USART_SendString(tb,0);

	int m = twi_master_read(DS1307_TWI_ADDR, DS1307_RAM_START, c, 10);	
	snprintf(tb, 41, "sr %d: %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X\n", m, c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7], c[8], c[9]);
	USART_SendString(tb,0);
	*/
//-------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------


//volatile uint32_t ds1307_sqw_counter = 0;	//rolls over every 6 days @ 8192 Hz counting
											//rolls over ever 18 hrs @ 65536 Hz counting
/*
uint32_t rtc_elapsed_time_msec( uint32_t t2, uint32_t t1)
//returns elapsed time in mentiseconds
{
	return ((uint32_t)(1000*((double)(t2-t1))/(2*DS1307_SQW_FREQ)));
}
*/

void rtc_init()
{
	if (!ds1307_is_init())
	{
		if (ds1307_init() < 0)
		{}//RTC error - do something appropriate
	}

	//ds1307_sqw_counter = 0;

	//turn on the PCINT9 interrupt (on PJ0)
	//PCMSK1 |= 0x02;
	//PCICR  |= 0x02;

}

/*
ISR(PCINT1_vect, ISR_BLOCK)
//incriments ds1307_sqw_counter on PCINT9
//(ie on both rising and falling edges)
//therefore it counts at 1/(2*DS1307_SQW_FREQ) resolution
{
	ds1307_sqw_counter++;
}
*/

/*
void msleep(uint32_t t)
//sleeps for t milliseconds
{
	volatile uint32_t start = ds1307_sqw_counter;

	while ( rtc_elapsed_time_msec(ds1307_sqw_counter, start) < t )
	{}

}
*/
#endif
